 /*
  * main.c - Monitor-netlink
  * 
  * This file makes use of the following monitor-netlink example:
  * 
  * https://github.com/lihongguang/monitor-netlink
  * 
  * All modifications to the original source file are:
  * Copyright (C) 2021 Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org>
  * 
  * simple-netaid is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
  * Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  * 
  * simple-netaid is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  * See the GNU General Public License for more details.
  * 
  * You should have received a copy of the GNU General Public License along
  * with this program.  If not, see <http://www.gnu.org/licenses/>.
  * 
  * See the COPYING file.
  */
   
  
#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <libgen.h>   // dirname() and basename()
#include <time.h>
#include <signal.h>
#include <memory.h>
#include <net/if.h>
#include <asm/types.h>
#include <arpa/inet.h>
#include <sys/un.h>      /*  UNIX socket  */
#include <sys/types.h>  
#include <sys/socket.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <netinet/in.h>
#include <assert.h>
#include <inttypes.h>
#include <sys/wait.h>
#include <ctype.h>      // isdigit()

#include <sys/wait.h>

#include <simple-netaid/sbuf.h>
#include <simple-netaid/iproute.h>
#include <simple-netaid/interfaces.h>
#include <simple-netaid/helpers.h>
#include <simple-netaid/ipaddr.h>
#include <simple-netaid/wireless.h>
#include <simple-netaid/ethlink.h>

#include "ini.h"
#include "def.h"
#include "config.h"
#include "util.h"
#include "sigact.h"
#include "policy.h"

#define MAX_SIZE 1024 

int DEBUG = 1;

int _DEBUG_MESSAGES = 0;
int _INFO_MESSAGES = 0;
int _WARN_MESSAGES = 1;
int _ERROR_MESSAGES = 1; 

// Global
const char *progname;

config_t *m_config = NULL;
const char *libpath;
char arg0[512] = {0};
int pid_fd = -1;      // lock file
struct ubus_context *ctx;
const char *ubus_socket="/var/run/ubus/ubus.sock";

// set debug level, by setting for info and debug messages
static
void set_debug_level(int d)
{   
   if(d >= 1) {
      _INFO_MESSAGES = 1;
   }
   if(d >= 2) {
      _DEBUG_MESSAGES = 1;
   }
}

// set error level, by setting for warning and error messages
static
void set_error_level(int e)
{ 
   if(e >= 1) {
      _ERROR_MESSAGES = 1;
   }
   if(e >= 2) {
      _WARN_MESSAGES = 1;
   }
}

int main (int argc, char **argv)
{  
   int rc __attribute__((unused)) = 0; 
   int count __attribute__((unused)) = 0;
   char output[256] = {0};
   char *cad = NULL;

   // arg0 -> the whole path
   strncpy_t(arg0, 512, argv[0], strlen(argv[0]));
   
   // cad -> the basename of arg0
   if ((cad = strrchr(argv[0], '/')))
      cad++;
   else
      cad = argv[0];
      
   progname = strdup_or_null(cad);
         
   m_config = (config_t*)calloc(sizeof(config_t), 1);
   if(m_config == NULL) {
      return -ENOMEM;
   }
   
   rc = config_init(m_config);
   if(rc != 0) {
      log_error("%s: config_init: %s\n", progname, strerror(errno));
      exit(EXIT_FAILURE);
   }
   
   // parse config options from command-line 
   rc = config_load_from_args(m_config, argc, argv);      
   if(rc != 0) {
      log_error("%s: snetaid_config_load_from_argv: %s\n", progname, strerror(errno));
      config_usage(argv[0]);
      return rc;
   }
   
   // if we didn't get a config file, use the default one
   if(m_config->config_path == NULL) {
      
      m_config->config_path = strdup_or_null(DEFAULT_CONFIG_FILE);
      if(m_config->config_path == NULL) {
         // OOM 
         return -ENOMEM;
      }
   }
   
   // load from file
   rc = config_load(m_config->config_path, m_config);
   if(rc != 0) {
      log_error("%s: config_load('%s'): %s\n", progname, m_config->config_path, strerror(errno));
      exit(EXIT_FAILURE);
   }

   if (!m_config->foreground) {
     /* It is also possible to use glibc function deamon()
      * at this point, but it is useful to customize your daemon. */
      daemonize();
   }
      
      /* Open system log and write message to it */
      openlog(progname, LOG_PID|LOG_CONS, LOG_DAEMON);
      log_info("Started %s", progname);
      
      sigaction_init();
      
      alarm(1);
      
      // HACK ALERT: whilst being an exotic case, it may happen that ubus and snetaid are 
      // being supervised by different init systems -say runit and sysvinit-, in which case 
      // the latter may start before the bus, contrary to expectations.
      while(count<10) {
         FILE *pf = popen("ps -A | grep ubusd", "r");
         if(!pf) {
            printf("Sorry, there was an error opening the pipe\n");
            exit(EXIT_FAILURE);
         }
   
         if(fgets(output, sizeof(output), pf) != NULL)
            break;
         else {
            sleep(1);
            count++;
         }
         pclose(pf);
      }   
            
      uloop_init();
     
      ctx=ubus_connect(ubus_socket);
      if(ctx==NULL) exit(EXIT_FAILURE);
 
      ubus_add_uloop(ctx);
        
      rc=ubus_add_object(ctx, &netaid_obj);
      if(rc!=0) {
         log_error("%s: Ubus failure.\n", progname, strerror(errno)); 
         ubus_free(ctx);
         exit(EXIT_FAILURE);
      }
      
      /* uloop routine: events monitoring and callback provoking */
      uloop_run();
      uloop_done();
   

      ubus_free(ctx);
      
   return rc;   
}
